# -*- coding: utf-8; -*-
################################################################################
#
#  Wutta-Continuum -- SQLAlchemy Versioning for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
SQLAlchemy-Continuum utilities
"""

import sqlalchemy as sa
from sqlalchemy import orm
import sqlalchemy_continuum as continuum


OPERATION_TYPES = {
    continuum.Operation.INSERT: "INSERT",
    continuum.Operation.UPDATE: "UPDATE",
    continuum.Operation.DELETE: "DELETE",
}


def render_operation_type(operation_type):
    """
    Render a SQLAlchemy-Continuum ``operation_type`` from a version
    record, for display to user.

    :param operation_type: Value of same name from a version record.
       Must be one of:

       * :attr:`sqlalchemy_continuum:sqlalchemy_continuum.operation.Operation.INSERT`
       * :attr:`sqlalchemy_continuum:sqlalchemy_continuum.operation.Operation.UPDATE`
       * :attr:`sqlalchemy_continuum:sqlalchemy_continuum.operation.Operation.DELETE`

    :returns: Display name for the operation type, as string.
    """
    return OPERATION_TYPES[operation_type]


def model_transaction_query(instance, session=None, model_class=None):
    """
    Make a query capable of finding all SQLAlchemy-Continuum
    ``transaction`` records associated with the given model instance.

    :param instance: Instance of a versioned :term:`data model`.

    :param session: Optional :term:`db session` to use for the query.
       If not specified, will be obtained from the ``instance``.

    :param model_class: Optional :term:`data model` class to query.
       If not specified, will be obtained from the ``instance``.

    :returns: SQLAlchemy query object.  Note that it will *not* have an
       ``ORDER BY`` clause yet.
    """
    if not session:
        session = orm.object_session(instance)
    if not model_class:
        model_class = type(instance)

    txncls = continuum.transaction_class(model_class)
    vercls = continuum.version_class(model_class)

    query = session.query(txncls).join(
        vercls,
        sa.and_(vercls.uuid == instance.uuid, vercls.transaction_id == txncls.id),
    )

    return query



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

from importlib.metadata import version as get_version

project = "Wutta-Continuum"
copyright = "2024, Lance Edgar"
author = "Lance Edgar"
release = get_version("Wutta-Continuum")

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinx_mapping = {
    "sqlalchemy": ("http://docs.sqlalchemy.org/en/latest/", None),
    "sqlalchemy-continuum": (
        "https://sqlalchemy-continuum.readthedocs.io/en/latest/",
        None,
    ),
    "wuttjamaican": ("https://docs.wuttaproject.org/wuttjamaican/", None),
}


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



"""remove password

Revision ID: 989392cc191d
Revises: 0a5f8ac0cd06
Create Date: 2025-10-29 19:42:52.985167

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "989392cc191d"
down_revision: Union[str, None] = "0a5f8ac0cd06"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # user
    op.drop_column("user_version", "password")


def downgrade() -> None:

    # user
    op.add_column(
        "user_version",
        sa.Column(
            "password", sa.VARCHAR(length=60), autoincrement=False, nullable=True
        ),
    )



# -*- coding: utf-8; -*-
################################################################################
#
#  Wutta-Continuum -- SQLAlchemy Versioning for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
App Provider
"""

from wuttjamaican.app import AppProvider


class WuttaContinuumAppProvider(AppProvider):
    """
    The :term:`app provider` for WuttaWeb.  This adds some methods to
    the :term:`app handler`, which are specific to Wutta-Continuum.
    """

    def continuum_is_enabled(self):
        """
        Returns boolean indicating if Wutta-Continuum is enabled.

        This checks the config value as described in
        :doc:`/narr/install`; default will be ``False``.
        """
        return self.config.get_bool(
            "wutta_continuum.enable_versioning", usedb=False, default=False
        )



"""first versioning tables

Revision ID: 71406251b8e7
Revises:
Create Date: 2024-08-27 18:28:31.488291

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import wuttjamaican.db.util


# revision identifiers, used by Alembic.
revision: str = "71406251b8e7"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = ("wutta_continuum",)
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # transaction
    op.create_table(
        "transaction",
        sa.Column("issued_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.BigInteger(), autoincrement=True, nullable=False),
        sa.Column("remote_addr", sa.String(length=50), nullable=True),
        sa.Column("user_id", wuttjamaican.db.util.UUID(), nullable=True),
        sa.ForeignKeyConstraint(
            ["user_id"], ["user.uuid"], name=op.f("fk_transaction_user_id_user")
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_transaction")),
    )
    op.create_index(
        op.f("ix_transaction_user_id"), "transaction", ["user_id"], unique=False
    )

    # person
    op.create_table(
        "person_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "full_name", sa.String(length=100), autoincrement=False, nullable=True
        ),
        sa.Column(
            "first_name", sa.String(length=50), autoincrement=False, nullable=True
        ),
        sa.Column(
            "middle_name", sa.String(length=50), autoincrement=False, nullable=True
        ),
        sa.Column(
            "last_name", sa.String(length=50), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_person_version")
        ),
    )
    op.create_index(
        op.f("ix_person_version_end_transaction_id"),
        "person_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_person_version_operation_type"),
        "person_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_person_version_transaction_id"),
        "person_version",
        ["transaction_id"],
        unique=False,
    )

    # user
    op.create_table(
        "user_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column("username", sa.String(length=25), autoincrement=False, nullable=True),
        sa.Column("password", sa.String(length=60), autoincrement=False, nullable=True),
        sa.Column(
            "person_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("active", sa.Boolean(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("uuid", "transaction_id", name=op.f("pk_user_version")),
    )
    op.create_index(
        op.f("ix_user_version_end_transaction_id"),
        "user_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_user_version_operation_type"),
        "user_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_user_version_transaction_id"),
        "user_version",
        ["transaction_id"],
        unique=False,
    )

    # role
    op.create_table(
        "role_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column("name", sa.String(length=100), autoincrement=False, nullable=True),
        sa.Column("notes", sa.Text(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("uuid", "transaction_id", name=op.f("pk_role_version")),
    )
    op.create_index(
        op.f("ix_role_version_end_transaction_id"),
        "role_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_role_version_operation_type"),
        "role_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_role_version_transaction_id"),
        "role_version",
        ["transaction_id"],
        unique=False,
    )

    # user_x_role
    op.create_table(
        "user_x_role_version",
        sa.Column(
            "uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "user_uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "role_uuid", wuttjamaican.db.util.UUID(), autoincrement=False, nullable=True
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "uuid", "transaction_id", name=op.f("pk_user_x_role_version")
        ),
    )
    op.create_index(
        op.f("ix_user_x_role_version_end_transaction_id"),
        "user_x_role_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_user_x_role_version_operation_type"),
        "user_x_role_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_user_x_role_version_transaction_id"),
        "user_x_role_version",
        ["transaction_id"],
        unique=False,
    )

    # permission
    op.create_table(
        "permission_version",
        sa.Column(
            "role_uuid",
            wuttjamaican.db.util.UUID(),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column(
            "permission", sa.String(length=254), autoincrement=False, nullable=False
        ),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint(
            "role_uuid",
            "permission",
            "transaction_id",
            name=op.f("pk_permission_version"),
        ),
    )
    op.create_index(
        op.f("ix_permission_version_end_transaction_id"),
        "permission_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_permission_version_operation_type"),
        "permission_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_permission_version_transaction_id"),
        "permission_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade() -> None:

    # permission
    op.drop_index(
        op.f("ix_permission_version_transaction_id"), table_name="permission_version"
    )
    op.drop_index(
        op.f("ix_permission_version_operation_type"), table_name="permission_version"
    )
    op.drop_index(
        op.f("ix_permission_version_end_transaction_id"),
        table_name="permission_version",
    )
    op.drop_table("permission_version")

    # user_x_role
    op.drop_index(
        op.f("ix_user_x_role_version_transaction_id"), table_name="user_x_role_version"
    )
    op.drop_index(
        op.f("ix_user_x_role_version_operation_type"), table_name="user_x_role_version"
    )
    op.drop_index(
        op.f("ix_user_x_role_version_end_transaction_id"),
        table_name="user_x_role_version",
    )
    op.drop_table("user_x_role_version")

    # role
    op.drop_index(op.f("ix_role_version_transaction_id"), table_name="role_version")
    op.drop_index(op.f("ix_role_version_operation_type"), table_name="role_version")
    op.drop_index(op.f("ix_role_version_end_transaction_id"), table_name="role_version")
    op.drop_table("role_version")

    # user
    op.drop_index(op.f("ix_user_version_transaction_id"), table_name="user_version")
    op.drop_index(op.f("ix_user_version_operation_type"), table_name="user_version")
    op.drop_index(op.f("ix_user_version_end_transaction_id"), table_name="user_version")
    op.drop_table("user_version")

    # person
    op.drop_index(op.f("ix_person_version_transaction_id"), table_name="person_version")
    op.drop_index(op.f("ix_person_version_operation_type"), table_name="person_version")
    op.drop_index(
        op.f("ix_person_version_end_transaction_id"), table_name="person_version"
    )
    op.drop_table("person_version")

    # transaction
    op.drop_index(op.f("ix_transaction_user_id"), table_name="transaction")
    op.drop_table("transaction")



# -*- coding: utf-8; -*-
"""
Tasks for Wutta-Continuum
"""

import os
import shutil

from invoke import task


@task
def release(c, skip_tests=False):
    """
    Release a new version of Wutta-Continuum
    """
    if not skip_tests:
        c.run("pytest")

    if os.path.exists("dist"):
        shutil.rmtree("dist")

    c.run("python -m build --sdist")
    c.run("twine upload dist/*")



# -*- coding: utf-8; -*-

from unittest import TestCase

import sqlalchemy_continuum as continuum

from wutta_continuum import util as mod
from wutta_continuum.testing import VersionTestCase


class TestRenderOperationType(TestCase):

    def test_basic(self):
        self.assertEqual(
            mod.render_operation_type(continuum.Operation.INSERT), "INSERT"
        )
        self.assertEqual(
            mod.render_operation_type(continuum.Operation.UPDATE), "UPDATE"
        )
        self.assertEqual(
            mod.render_operation_type(continuum.Operation.DELETE), "DELETE"
        )


class TestModelTransactionQuery(VersionTestCase):

    def test_basic(self):
        model = self.app.model

        user = model.User(username="fred")
        self.session.add(user)
        self.session.commit()

        query = mod.model_transaction_query(user)
        self.assertEqual(query.count(), 1)
        txn = query.one()

        UserVersion = continuum.version_class(model.User)
        version = self.session.query(UserVersion).one()
        self.assertIs(version.transaction, txn)



# -*- coding: utf-8; -*-
################################################################################
#
#  Wutta-Continuum -- SQLAlchemy Versioning for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
App Configuration
"""

import socket

from sqlalchemy.orm import configure_mappers
from sqlalchemy_continuum import make_versioned
from sqlalchemy_continuum.plugins import Plugin

from wuttjamaican.conf import WuttaConfigExtension
from wuttjamaican.util import load_object


class WuttaContinuumConfigExtension(WuttaConfigExtension):
    """
    App :term:`config extension` for Wutta-Continuum.

    This adds a startup hook, which can optionally turn on the
    SQLAlchemy-Continuum versioning features for the main app DB.
    """

    key = "wutta_continuum"

    def startup(self, config):  # pylint: disable=empty-docstring
        """
        Perform final configuration setup for app startup.

        This will do nothing at all, unless config enables the
        versioning feature.  This must be done in config file and not
        in DB settings table:

        .. code-block:: ini

           [wutta_continuum]
           enable_versioning = true

        Once enabled, this method will configure the integration, via
        these steps:

        1. call :func:`sqlalchemy-continuum:sqlalchemy_continuum.make_versioned()`
        2. call :meth:`wuttjamaican:wuttjamaican.app.AppHandler.get_model()`
        3. call :func:`sqlalchemy:sqlalchemy.orm.configure_mappers()`

        For more about SQLAlchemy-Continuum see
        :doc:`sqlalchemy-continuum:intro`.
        """
        # only do this if config enables it
        if not config.get_bool(
            "wutta_continuum.enable_versioning", usedb=False, default=False
        ):
            return

        # create wutta plugin, to assign user and ip address
        spec = config.get(
            "wutta_continuum.wutta_plugin_spec",
            usedb=False,
            default="wutta_continuum.conf:WuttaContinuumPlugin",
        )
        plugin = load_object(spec)

        app = config.get_app()
        if "model" in app.__dict__:
            raise RuntimeError("something not right, app already has model")

        # let sqlalchemy-continuum do its thing
        make_versioned(plugins=[plugin()])

        # must load model *between* prev and next calls
        app.get_model()

        # let sqlalchemy do its thing
        configure_mappers()


class WuttaContinuumPlugin(Plugin):
    """
    SQLAlchemy-Continuum manager plugin for Wutta-Continuum.

    This tries to assign the current user and IP address to the
    transaction.

    It will assume the "current machine" IP address, which may be
    suitable for some apps but not all (e.g. web apps, where IP
    address should reflect an arbitrary client machine).

    However it does not actually have a way to determine the current
    user.  WuttaWeb therefore uses a different plugin, based on this
    one, to get both the user and IP address from current request.

    You can override this to use a custom plugin for this purpose; if
    so you must specify in your config file:

    .. code-block:: ini

       [wutta_continuum]
       wutta_plugin_spec = poser.db.continuum:PoserContinuumPlugin

    See also the SQLAlchemy-Continuum docs for
    :doc:`sqlalchemy-continuum:plugins`.
    """

    def get_remote_addr(  # pylint: disable=empty-docstring,unused-argument
        self, uow, session
    ):
        """ """
        host = socket.gethostname()
        return socket.gethostbyname(host)

    def get_user_id(  # pylint: disable=empty-docstring,unused-argument
        self, uow, session
    ):
        """ """
        return None

    def transaction_args(self, uow, session):  # pylint: disable=empty-docstring
        """ """
        kwargs = {}

        remote_addr = self.get_remote_addr(uow, session)
        if remote_addr:
            kwargs["remote_addr"] = remote_addr

        user_id = self.get_user_id(uow, session)  # pylint: disable=assignment-from-none
        if user_id:
            kwargs["user_id"] = user_id

        return kwargs



# -*- coding: utf-8; -*-

from wuttjamaican.testing import DataTestCase

from wutta_continuum import app as mod


class TestWuttaContinuumAppProvider(DataTestCase):

    def make_provider(self):
        return mod.WuttaContinuumAppProvider(self.config)

    def test_continuum_is_enabled(self):

        # off by default
        provider = self.make_provider()
        self.assertFalse(provider.continuum_is_enabled())

        # but can be turned on
        self.config.setdefault("wutta_continuum.enable_versioning", "true")
        self.assertTrue(provider.continuum_is_enabled())



"""add user.prevent_edit

Revision ID: 0a5f8ac0cd06
Revises: 71406251b8e7
Create Date: 2024-11-24 17:39:57.415425

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = "0a5f8ac0cd06"
down_revision: Union[str, None] = "71406251b8e7"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:

    # user
    op.add_column(
        "user_version",
        sa.Column("prevent_edit", sa.Boolean(), autoincrement=False, nullable=True),
    )


def downgrade() -> None:

    # user
    op.drop_column("user_version", "prevent_edit")



# -*- coding: utf-8; -*-
################################################################################
#
#  Wutta-Continuum -- SQLAlchemy Versioning for Wutta Framework
#  Copyright © 2024-2025 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Testing utilities
"""

import sys

import sqlalchemy_continuum as continuum

from wuttjamaican.testing import DataTestCase

from wutta_continuum.conf import WuttaContinuumConfigExtension


class VersionTestCase(DataTestCase):
    """
    Base class for test suites requiring the SQLAlchemy-Continuum
    versioning feature.

    This inherits from
    :class:`~wuttjamaican:wuttjamaican.testing.DataTestCase`.
    """

    def setUp(self):
        self.setup_versioning()

    def setup_versioning(self):
        """
        Do setup tasks relating to this class, as well as its parent(s):

        * call :meth:`wuttjamaican:wuttjamaican.testing.DataTestCase.setup_db()`

          * this will in turn call :meth:`make_config()`
        """
        self.setup_db()

    def tearDown(self):
        self.teardown_versioning()

    def teardown_versioning(self):
        """
        Do teardown tasks relating to this class, as well as its parent(s):

        * call :func:`sqlalchemy-continuum:sqlalchemy_continuum.remove_versioning()`
        * call :meth:`wuttjamaican:wuttjamaican.testing.DataTestCase.teardown_db()`
        """
        continuum.remove_versioning()
        continuum.versioning_manager.transaction_cls = continuum.TransactionFactory()
        self.teardown_db()

    def make_config(self, **kwargs):
        """
        Make and customize the config object.

        We override this to explicitly enable the versioning feature.
        """
        config = super().make_config(**kwargs)
        config.setdefault("wutta_continuum.enable_versioning", "true")

        # nb. must purge model classes from sys.modules, so they will
        # be reloaded and sqlalchemy-continuum can reconfigure
        if "wuttjamaican.db.model" in sys.modules:
            del sys.modules["wuttjamaican.db.model.batch"]
            del sys.modules["wuttjamaican.db.model.upgrades"]
            del sys.modules["wuttjamaican.db.model.auth"]
            del sys.modules["wuttjamaican.db.model.base"]
            del sys.modules["wuttjamaican.db.model"]

        ext = WuttaContinuumConfigExtension()
        ext.startup(config)
        return config



# -*- coding: utf-8; -*-

import socket

from unittest.mock import patch

from wuttjamaican.testing import ConfigTestCase, DataTestCase

from wutta_continuum import conf as mod


class TestWuttaContinuumConfigExtension(ConfigTestCase):

    def make_extension(self):
        return mod.WuttaContinuumConfigExtension()

    def test_startup_without_versioning(self):
        ext = self.make_extension()
        with patch.object(mod, "make_versioned") as make_versioned:
            with patch.object(mod, "configure_mappers") as configure_mappers:
                ext.startup(self.config)
                make_versioned.assert_not_called()
                configure_mappers.assert_not_called()

    def test_startup_with_versioning(self):
        ext = self.make_extension()
        with patch.object(mod, "make_versioned") as make_versioned:
            with patch.object(mod, "configure_mappers") as configure_mappers:
                self.config.setdefault("wutta_continuum.enable_versioning", "true")
                ext.startup(self.config)
                make_versioned.assert_called_once()
                configure_mappers.assert_called_once_with()

    def test_startup_with_error(self):
        ext = self.make_extension()
        with patch.object(mod, "make_versioned") as make_versioned:
            with patch.object(mod, "configure_mappers") as configure_mappers:
                self.config.setdefault("wutta_continuum.enable_versioning", "true")
                # nb. it is an error for the model to be loaded prior to
                # calling make_versioned() for sqlalchemy-continuum
                self.app.get_model()
                self.assertRaises(RuntimeError, ext.startup, self.config)
                make_versioned.assert_not_called()
                configure_mappers.assert_not_called()


class TestWuttaContinuumPlugin(DataTestCase):

    def make_plugin(self):
        return mod.WuttaContinuumPlugin()

    def test_remote_addr(self):
        plugin = self.make_plugin()
        with patch.object(socket, "gethostbyname", return_value="127.0.0.1"):
            self.assertEqual(plugin.get_remote_addr(None, self.session), "127.0.0.1")

    def test_user_id(self):
        plugin = self.make_plugin()
        self.assertIsNone(plugin.get_user_id(None, self.session))

    def test_transaction_args(self):
        plugin = self.make_plugin()
        with patch.object(socket, "gethostbyname", return_value="127.0.0.1"):
            self.assertEqual(
                plugin.transaction_args(None, self.session),
                {"remote_addr": "127.0.0.1"},
            )

            with patch.object(plugin, "get_user_id", return_value="some-random-uuid"):
                self.assertEqual(
                    plugin.transaction_args(None, self.session),
                    {"remote_addr": "127.0.0.1", "user_id": "some-random-uuid"},
                )
